import numpy as np
from scipy.integrate import solve_ivp
from numba import njit
from jitr import (
    ProjectileTargetSystem,
    InteractionMatrix,
    ChannelData,
    LagrangeRMatrixSolver,
    woods_saxon_potential,
    coulomb_charged_sphere,
    delta,
    smatrix,
    schrodinger_eqn_ivp_order1,
)


@njit
def interaction(r, *args):
    (V0, W0, R0, a0, zz, r_c) = args
    return woods_saxon_potential(r, V0, W0, R0, a0) + coulomb_charged_sphere(r, zz, r_c)


def rmse_RK_LM(nwaves: int = 5):
    r"""Test with simple Woods-Saxon plus coulomb without spin-orbit coupling"""

    lgrid = np.arange(0, nwaves - 1, 1)
    egrid = np.linspace(0.01, 100, 10)
    nodes_within_radius = 5

    # channels are the same except for l and uncoupled
    # so just set up a single channel system. We will set
    # incident energy and l later
    sys = ProjectileTargetSystem(
        np.array([939.0]),
        np.array([nodes_within_radius * (2 * np.pi)]),
        l=np.array([0]),
        Ztarget=40,
        Zproj=1,
        nchannels=1,
    )

    # Lagrange-Mesh solver, don't set the energy
    solver_lm = LagrangeRMatrixSolver(40, 1, sys, ecom=None)

    # Woods-Saxon potential parameters
    V0 = 60  # real potential strength
    W0 = 20  # imag potential strength
    R0 = 4  # Woods-Saxon potential radius
    a0 = 0.5  # Woods-Saxon potential diffuseness
    RC = R0  # Coulomb cutoff

    params = (V0, W0, R0, a0, sys.Zproj * sys.Ztarget, RC)

    # use same interaction for all channels
    interaction_matrix = InteractionMatrix(1)
    interaction_matrix.set_local_interaction(interaction, args=params)

    error_matrix = np.zeros((len(lgrid), len(egrid)), dtype=complex)

    for i, e in enumerate(egrid):
        for l in lgrid:
            sys.l = np.array([l])
            ch = sys.build_channels(e)
            a = ch[0].domain[1]

            # Runge-Kutta
            sol_rk = solve_ivp(
                lambda s, y,: schrodinger_eqn_ivp_order1(
                    s, y, ch[0], interaction_matrix.local_matrix[0, 0], params
                ),
                ch[0].domain,
                ch[0].initial_conditions(),
                dense_output=True,
                atol=1.0e-12,
                rtol=1.0e-9,
            ).sol

            R_rk = sol_rk(a)[0] / (a * sol_rk(a)[1])
            S_rk = smatrix(R_rk, a, l, ch[0].eta)

            # Lagrange-Legendre R-Matrix
            R_lm, S_lm, uext_boundary = solver_lm.solve(interaction_matrix, ch, ecom=e)

            # comparison between solvers
            delta_lm, atten_lm = delta(S_lm)
            delta_rk, atten_rk = delta(S_rk)

            err = 0 + 0j

            if np.fabs(delta_rk) > 1e-12:
                err += np.fabs(delta_lm - delta_rk)

            if np.fabs(atten_rk) > 1e-12:
                err += 1j * np.fabs(atten_lm - atten_rk)

            error_matrix[l, i] = err

        return np.real(error_matrix), np.imag(error_matrix)


def test_local():
    rtol = 1.0e-2
    real_err, imag_err = rmse_RK_LM()
    assert not np.any(real_err / 180 > rtol)
    assert not np.any(imag_err / 180 > rtol)



import numpy as np
from matplotlib import pyplot as plt

from jitr import (
    ProjectileTargetSystem,
    InteractionMatrix,
    ChannelData,
    Wavefunctions,
    LagrangeRMatrixSolver,
    yamaguchi_potential,
    yamaguchi_swave_delta,
    delta,
    smatrix,
    hbarc,
)


def nonlocal_interaction_example():
    alpha = 0.2316053  # fm**-1
    beta = 1.3918324  # fm**-1
    W0 = 41.472  # Mev fm**2

    params = (W0, beta, alpha)
    mass = hbarc**2 / (2 * W0)

    ecom = 0.1

    sys = ProjectileTargetSystem(
        reduced_mass=np.array([mass]),
        channel_radii=np.array([30.0]),
        l=np.array([0], dtype=np.int64),
    )
    channels = sys.build_channels(ecom)

    interaction_matrix = InteractionMatrix(1)
    interaction_matrix.set_nonlocal_interaction(yamaguchi_potential, args=params)

    solver = LagrangeRMatrixSolver(100, 1, sys, ecom=ecom, channel_matrix=channels)
    _, S, _ = solver.solve(interaction_matrix, channels)

    delta = np.rad2deg(np.real(np.log(S[0, 0]) / 2j))

    print("\nYamaguchi potential test:\n delta:")
    print("Lagrange-Legendre Mesh: {:.6f} [degrees]".format(delta))
    print(
        "Analytic              : {:.6f} [degrees]".format(
            yamaguchi_swave_delta(channels[0].k, *params)
        )
    )


if __name__ == "__main__":
    nonlocal_interaction_example()



import numpy as np
from numba.experimental import jitclass
from numba import int32, float64, njit

from .system import InteractionMatrix
from .channel import ChannelData
from .utils import eval_scaled_interaction, eval_scaled_nonlocal_interaction, block


spec = [
    ("nbasis", int32),
    ("nchannels", int32),
    ("abscissa", float64[:]),
    ("weights", float64[:]),
]


@jitclass(spec)
class LagrangeRMatrixKernel:
    r"""
    Lagrange-Legendre mesh for the Bloch-Schroedinger equation following:
    Baye, D. (2015). The Lagrange-mesh method. Physics reports, 565, 1-107,
    with the only difference being the domain is scaled in each channel; e.g.
    r -> s_i = r * k_i, and each channel's equation is then divided by it's
    asymptotic kinetic energy in the channel T_i = E_inc - E_i
    """

    def __init__(self, nbasis, nchannels, abscissa, weights):
        """
        Constructs the Bloch-Schroedinger equation in a basis of nbasis
        Lagrange-Legendre functions shifted and scaled onto [0,k*a] and regulated by 1/k*r,
        and solved by direct matrix inversion.

        """
        self.nbasis = nbasis
        self.nchannels = nchannels
        self.abscissa = abscissa
        self.weights = weights

    def interaction_matrix_element(
        self,
        n,
        m,
        ch: ChannelData,
        local_interaction,
        nonlocal_interaction=None,
        args_local=(),
        args_nonlocal=(),
    ):
        Vnm = self.local_potential_matrix_element(
            n, m, local_interaction, ch, args_local
        )
        if nonlocal_interaction is not None:
            Vnm += self.nonlocal_potential_matrix_element(
                n, m, nonlocal_interaction, ch, args_nonlocal
            )
        return Vnm

    def local_potential_matrix_element(
        self, n, m, interaction, ch: ChannelData, args=()
    ):
        """
        evaluates the (n,m)th matrix element for the given local interaction
        """
        assert n <= self.nbasis and n >= 1
        assert m <= self.nbasis and m >= 1

        if n != m:
            return 0  # local potentials are diagonal

        xn = self.abscissa[n - 1]

        a = ch.domain[1]
        s = xn * a

        return eval_scaled_interaction(s, interaction, ch, args)

    def nonlocal_potential_matrix_element(self, n, m, interaction, ch, args=()):
        """
        evaluates the (n,m)th matrix element for the given non-local interaction
        """
        assert n <= self.nbasis and n >= 1
        assert m <= self.nbasis and m >= 1

        xn = self.abscissa[n - 1]
        xm = self.abscissa[m - 1]
        wn = self.weights[n - 1]
        wm = self.weights[m - 1]

        a = ch.domain[1]

        s = xn * a
        sp = xm * a

        utilde = eval_scaled_nonlocal_interaction(s, sp, interaction, ch, args)

        return utilde * np.sqrt(wm * wn) * a

    def kinetic_bloch_matrix_element(self, n, m, ch):
        """
        evaluates the (n,m)th matrix element for the kinetic energy + Bloch operator
        """
        assert n <= self.nbasis and n >= 1
        assert m <= self.nbasis and m >= 1

        xn, xm = self.abscissa[n - 1], self.abscissa[m - 1]
        N = self.nbasis

        a = ch.domain[1]
        l = ch.l

        if n == m:
            centrifugal = l * (l + 1) / (a * xn) ** 2
            # Eq. 3.128 in [Baye, 2015], scaled by 1/E and with r->s=kr
            return ((4 * N**2 + 4 * N + 3) * xn * (1 - xn) - 6 * xn + 1) / (
                3 * xn**2 * (1 - xn) ** 2
            ) / a**2 + centrifugal
        else:
            # Eq. 3.129 in [Baye, 2015], scaled by 1/E and with r->s=kr
            return (
                (-1.0) ** (n + m)
                * (
                    (N**2 + N + 1.0)
                    + (xn + xm - 2 * xn * xm) / (xn - xm) ** 2
                    - 1.0 / (1.0 - xn)
                    - 1.0 / (1.0 - xm)
                )
                / np.sqrt(xn * xm * (1.0 - xn) * (1.0 - xm))
                / a**2
            )

    def single_channel_free_matrix(self, ch: ChannelData):
        r"""
        Returns the Free Bloch-Schroedinger equation in Lagrange-Legendre coordinates as
        an upper-triangular matrix (the full matrix is symmetric).
        """
        F = np.zeros((self.nbasis, self.nbasis), dtype=np.complex128)
        for n in range(1, self.nbasis + 1):
            for m in range(n, self.nbasis + 1):
                F[n - 1, m - 1] = self.kinetic_bloch_matrix_element(n, m, ch)

        F -= np.diag(np.ones(self.nbasis))
        F = F + np.triu(F, k=1).T

        return F

    def single_channel_bloch_se_matrix(
        self,
        i,
        j,
        local_interaction,
        nonlocal_interaction,
        is_symmetric,
        ch: ChannelData,
        free_matrix: np.array = None,
        args_local=None,
        args_nonlocal=None,
    ):
        r"""Implements Eq. 6.10 in [Baye, 2015], scaled by 1/E and with r->s=kr. The diagonal
        submatrices in channel space include full bloch-SE.
        """
        # Free matrix (Kinetic and Bloch terms) only nonzero on channel diagonal
        if i == j:
            # if channel free matrix has been precomputed for us, use it;
            if free_matrix is not None:
                F = free_matrix
            else:
                # otherwise calculate it
                F = self.single_channel_free_matrix(ch)
        else:
            F = np.zeros((self.nbasis, self.nbasis), dtype=np.complex128)

        # calculate interaction matrix
        V = np.zeros_like(F, dtype=np.complex128)

        # calculate each upper tri interaction matrix elem for channel
        for n in range(1, self.nbasis + 1):
            for m in range(n, self.nbasis + 1):
                Vnm = self.interaction_matrix_element(
                    n,
                    m,
                    ch,
                    local_interaction,
                    nonlocal_interaction,
                    args_local,
                    args_nonlocal,
                )
                V[n - 1, m - 1] = Vnm

        if is_symmetric:
            # copy upper tri to lower
            V = V + np.triu(V, k=1).T
        else:
            # calculate lower tri matrix elements
            for n in range(1, self.nbasis + 1):
                for m in range(n + 1, self.nbasis + 1):
                    Vmn = self.interaction_matrix_element(
                        m,
                        n,
                        ch,
                        local_interaction,
                        nonlocal_interaction,
                        args_local,
                        args_nonlocal,
                    )
                    V[m - 1, n - 1] = Vmn

        # sum free and potential matrices
        return V + F


@njit
def rmsolve_smatrix(
    A: np.array,
    b: np.array,
    asymptotics: tuple,
    incoming_weights: np.array,
    a: np.array,
    nchannels: np.int32,
    nbasis: np.int32,
):
    """
    Returns the multichannel R-Matrix, S-matrix, and wavefunction coefficients, all in Lagrange-
    Legendre coordinates, as well as the derivative of asymptotic channel Wavefunctions evaluated
    at the channel radius. Everything returned as block-matrices and block vectors in channel space.
    Uses `numpy.linalg.solve`, which should be faster than alternatives (e.g. scipy) for mid-size
    (n<100) Hermitian matrices. Note: speed will be dependent on numpy backend for linear algebra (MKL
    general solve for n<100 is typically faster than LAPACK hermitian solve).

    This follows:
    Descouvemont, P. (2016).
    An R-matrix package for coupled-channel problems in nuclear physics.
    Computer physics communications, 200, 199-219.
    """
    (Hp, Hm, Hpp, Hmp) = asymptotics

    # Eqn 15 in Descouvemont, 2016
    x = np.linalg.solve(A, b).reshape(nchannels, nbasis)
    R = x @ b.reshape(nchannels, nbasis).T / np.outer(a, a)

    # Eqn 17 in Descouvemont, 2016
    Zp = np.diag(Hp) - R * Hpp[:, np.newaxis] * a[:, np.newaxis]
    Zm = np.diag(Hm) - R * Hmp[:, np.newaxis] * a[:, np.newaxis]

    # Eqn 16 in Descouvemont, 2016
    S = np.linalg.solve(Zp, Zm)

    uext_prime_boundary = Hmp * incoming_weights - S @ Hpp

    return R, S, uext_prime_boundary


@njit
def rmsolve_wavefunction(
    A: np.array,
    b: np.array,
    asymptotics: tuple,
    incoming_weights: np.array,
    a: np.array,
    nchannels: np.int32,
    nbasis: np.int32,
):
    """
    Returns the multichannel R-Matrix, S-matrix, and wavefunction coefficients, all in Lagrange-
    Legendre coordinates, as well as the derivative of asymptotic channel Wavefunctions evaluated
    at the channel radius. Everything returned as block-matrices and block vectors in channel space.
    Uses `numpy.linalg.solve`, which should be faster than alternatives (e.g. scipy) for mid-size
    (n<100) Hermitian matrices. Note: speed will be dependent on numpy backend for linear algebra (MKL
    general solve for n<100 is typically faster than LAPACK hermitian solve).

    This follows:
    Descouvemont, P. (2016).
    An R-matrix package for coupled-channel problems in nuclear physics.
    Computer physics communications, 200, 199-219.
    and
    P. Descouvemont and D. Baye 2010 Rep. Prog. Phys. 73 036301
    """
    # TODO should we factorize A so we don't have to solve from scratch twice?
    R, S, uext_prime_boundary = rmsolve_smatrix(
        A, b, asymptotics, incoming_weights, a, nchannels, nbasis
    )

    # Eqn 3.92 in Descouvemont & Baye, 2010
    b = (b.reshape(nchannels, nbasis) * uext_prime_boundary[:, np.newaxis]).reshape(
        nchannels * nbasis
    )
    x = np.linalg.solve(A, b).reshape(nchannels, nbasis)

    return R, S, x, uext_prime_boundary



from .potentials import *

# from .utils import delta, smatrix, schrodinger_eqn_ivp_order1
from .utils import *
from .system import ProjectileTargetSystem, InteractionMatrix
from .channel import ChannelData, Wavefunctions
from .rmatrix_solver import LagrangeRMatrixSolver
from .rmatrix_kernel import LagrangeRMatrixKernel
from .__version__ import __version__



from numba.experimental import jitclass
from numba import float64, int64
import numpy as np

from .utils import (
    Gamow_factor,
    H_plus,
    H_minus,
    alpha,
    hbarc,
    eval_scaled_interaction,
    null,
    CoulombAsymptotics,
)

spec_ch = [
    ("l", int64),
    ("mass", float64),
    ("E", float64),
    ("k", float64),
    ("eta", float64),
    ("domain", float64[:]),
]


@jitclass(spec_ch)
class ChannelData:
    """ """

    def __init__(
        self,
        l: np.int64,
        reduced_mass: np.float64,
        a: np.float64,
        E: np.float64,
        k: np.float64,
        eta: np.float64,
    ):
        """ """
        self.l = l
        self.mass = reduced_mass
        self.E = E
        self.k = k
        self.eta = eta
        self.domain = np.array([1.0e-3, a], dtype=np.float64)

    def initial_conditions(self):
        """
        initial conditions for numerical integration in coordinate (s) space
        """
        s_0 = self.domain[0]
        C_l = Gamow_factor(self.l, self.eta)
        u0 = C_l * s_0 ** (self.l + 1)
        uprime0 = C_l * (self.l + 1) * s_0**self.l
        return np.array([u0 * (1 + 0j), uprime0 * (1 + 0j)])

    def s_grid(self, size=200):
        return np.linspace(self.domain[0], self.domain[1], size)


class Wavefunctions:
    """
    Represents a wavefunction, expressed internally to the channel as a linear combination
    of Lagrange-Legendre functions, and externally as a linear combination of incoming and
    outgoing Coulomb scattering wavefunctions
    """

    def __init__(
        self,
        solver,
        coeffs,
        S,
        uext_prime_boundary,
        incoming_weights,
        channels,
        asym=CoulombAsymptotics,
    ):
        self.solver = solver
        self.coeffs = coeffs
        self.S = S
        self.channels = channels
        self.incoming_weights = incoming_weights
        self.uext_prime_boundary = uext_prime_boundary
        self.asym = asym

    def uext(self):
        r"""Returns a callable which evaluates the asymptotic form of wavefunction in each channel,
        valid external to the channel radii. Follows Eqn. 3.79 from P. Descouvemont and D. Baye
        2010 Rep. Prog. Phys. 73 036301
        """

        def uext_channel(i):
            # TODO optimize this
            l = self.channels[i].l
            eta = self.channels[i].eta
            asym_func_in = lambda s: (
                self.incoming_weights[i] * H_minus(s, l, eta, asym=self.asym)
            )
            asym_func_out = lambda s: np.sum(
                [
                    self.incoming_weights[j]
                    * self.S[i, j]
                    * H_plus(s, l, eta, asym=self.asym)
                    for j in self.solver.kernel.nchannels
                ],
                axis=0,
            )
            return lambda s: asym_func_in(s) + asym_func_out(s)

        uext = []
        for i in range(self.solver.kernel.nchannels):
            uext.append(uext_channel(i))

        return uext

    def uint(self):
        def uint_channel(i):
            return lambda s: np.sum(
                [
                    self.coeffs[i, n] * self.solver.f(n + 1, i, s)
                    for n in range(self.solver.kernel.nbasis)
                ],
                axis=0,
            )

        uint = []
        for i in range(self.solver.kernel.nchannels):
            uint.append(uint_channel(i))

        return uint



import numpy as np
import scipy.special as sc
import scipy.linalg as la

from .utils import (
    hbarc,
    c,
    CoulombAsymptotics,
    FreeAsymptotics,
    H_plus,
    H_minus,
    H_plus_prime,
    H_minus_prime,
)
from .system import ProjectileTargetSystem, InteractionMatrix
from .rmatrix_kernel import LagrangeRMatrixKernel, rmsolve_smatrix, rmsolve_wavefunction


class LagrangeRMatrixSolver:
    def __init__(
        self,
        nbasis,
        nchannels,
        sys: ProjectileTargetSystem,
        ecom=None,
        channel_matrix=None,
        asym=None,
    ):
        r"""
        Parameters:
            a : channel radii

        """
        x, w = np.polynomial.legendre.leggauss(nbasis)
        abscissa = 0.5 * (x + 1)
        weights = 0.5 * w
        self.kernel = LagrangeRMatrixKernel(nbasis, nchannels, abscissa, weights)

        if asym is None:
            if sys.Zproj * sys.Ztarget > 0:
                asym = CoulombAsymptotics
            else:
                asym = FreeAsymptotics
        self.asym = asym

        self.sys = sys
        self.ecom = ecom
        if ecom is not None:
            self.set_energy(ecom)

        self.free_matrices = None
        if channel_matrix is not None:
            self.precompute_free_matrices(channel_matrix)

    def precompute_asymptotics(self, a, l, eta):
        # precompute asymptotic values of Lagrange-Legendre for each channel
        b = np.hstack(
            [
                [self.f(n, i, a[i]) for n in range(1, self.kernel.nbasis + 1)]
                for i in range(self.kernel.nchannels)
            ],
            dtype=np.complex128,
        )

        # precompute asymoptotic wavefunction and derivative in each channel
        Hp = np.array(
            [H_plus(ai, li, etai, asym=self.asym) for (ai, li, etai) in zip(a, l, eta)],
            dtype=np.complex128,
        )
        Hm = np.array(
            [
                H_minus(ai, li, etai, asym=self.asym)
                for (ai, li, etai) in zip(a, l, eta)
            ],
            dtype=np.complex128,
        )
        Hpp = np.array(
            [
                H_plus_prime(ai, li, etai, asym=self.asym)
                for (ai, li, etai) in zip(a, l, eta)
            ],
            dtype=np.complex128,
        )
        Hmp = np.array(
            [
                H_minus_prime(ai, li, etai, asym=self.asym)
                for (ai, li, etai) in zip(a, l, eta)
            ],
            dtype=np.complex128,
        )
        asymptotics = (Hp, Hm, Hpp, Hmp)
        return b, asymptotics

    def precompute_free_matrices(self, channel_matrix: np.array):
        nb = self.kernel.nbasis
        sz = nb * self.kernel.nchannels
        self.free_matrices = []
        for i in range(self.kernel.nchannels):
            self.free_matrices.append(
                self.kernel.single_channel_free_matrix(channel_matrix[i])
            )

    def set_energy(self, ecom: np.float64):
        r"""update precomputed values for new energy"""
        self.ecom = ecom
        self.b, self.boundary_asymptotic_wf = self.precompute_asymptotics(
            self.sys.channel_radii, self.sys.l, self.sys.eta(ecom)
        )

    def f(self, n, i, s):
        """
        nth basis function in channel i - Lagrange-Legendre polynomial of degree n shifted onto
        [0,a_i] and regularized by s/( a_i * xn)
        Note: n is indexed from 1 (constant function is not part of basis)
        """
        assert n <= self.kernel.nbasis and n >= 1

        x = s / self.sys.channel_radii[i]
        xn = self.kernel.abscissa[n - 1]

        # Eqn 3.122 in [Baye, 2015], with s = kr
        return (
            (-1.0) ** (self.kernel.nbasis - n)
            * np.sqrt((1 - xn) / xn)
            * sc.eval_legendre(self.kernel.nbasis, 2.0 * x - 1.0)
            * x
            / (x - xn)
        )

    def bloch_se_matrix(
        self,
        interaction_matrix: InteractionMatrix,
        channel_matrix: np.array,
    ):
        r"""Constructs the Bloch-Schrodinger equation in the Lagrange-Legendre basis"""
        nb = self.kernel.nbasis
        sz = nb * self.kernel.nchannels
        C = np.zeros((sz, sz), dtype=np.complex128)
        for i in range(self.kernel.nchannels):
            fi = None if self.free_matrices is None else self.free_matrices[i]
            for j in range(self.kernel.nchannels):
                args_local = interaction_matrix.local_args[i, j]
                args_nonlocal = interaction_matrix.nonlocal_args[i, j]
                C[i * nb : i * nb + nb, j * nb : j * nb + nb] = (
                    self.kernel.single_channel_bloch_se_matrix(
                        i,
                        j,
                        interaction_matrix.local_matrix[i, j],
                        interaction_matrix.nonlocal_matrix[i, j],
                        interaction_matrix.nonlocal_symmetric[i, j],
                        channel_matrix[i],
                        fi,
                        args_local,
                        args_nonlocal,
                    )
                )
        return C

    def solve(
        self,
        interaction_matrix: InteractionMatrix,
        channel_matrix: np.array,
        ecom=None,
        wavefunction=None,
    ):
        if ecom is not None:
            self.ecom = ecom
            self.set_energy(ecom)

        # either an ecom must be passed in, or the asymptotics must
        # have been pre-computed at some point using set_energy
        assert self.ecom is not None

        A = self.bloch_se_matrix(interaction_matrix, channel_matrix)

        if wavefunction is None:
            return rmsolve_smatrix(
                A,
                self.b,
                self.boundary_asymptotic_wf,
                self.sys.incoming_weights,
                self.sys.channel_radii,
                self.kernel.nchannels,
                self.kernel.nbasis,
            )
        else:
            return rmsolve_wavefunction(
                A,
                self.b,
                self.boundary_asymptotic_wf,
                self.sys.incoming_weights,
                self.sys.channel_radii,
                self.kernel.nchannels,
                self.kernel.nbasis,
            )



import numpy as np
from matplotlib import pyplot as plt
from scipy.integrate import solve_ivp
from numba import njit
from jitr import (
    ProjectileTargetSystem,
    InteractionMatrix,
    ChannelData,
    Wavefunctions,
    LagrangeRMatrixSolver,
    woods_saxon_potential,
    coulomb_charged_sphere,
    delta,
    smatrix,
    schrodinger_eqn_ivp_order1,
)


@njit
def interaction(r, *params):
    (V0, W0, R0, a0, zz, RC) = params
    return woods_saxon_potential(r, V0, W0, R0, a0) + coulomb_charged_sphere(r, zz, RC)


def channel_radius_dependence_test():
    E = 14.1
    nodes_within_radius = 5

    # Potential parameters
    V0 = 60  # real potential strength
    W0 = 20  # imag potential strength
    R0 = 4  # Woods-Saxon potential radius
    a0 = 0.5  # Woods-Saxon potential diffuseness
    params = (V0, W0, R0, a0)

    a_grid = np.linspace(10, 30, 50)
    delta_grid = np.zeros_like(a_grid, dtype=complex)

    ints = InteractionMatrix(1)
    ints.set_local_interaction(woods_saxon_potential, args=params)

    for i, a in enumerate(a_grid):
        sys = ProjectileTargetSystem(
            reduced_mass=np.array([939.0], dtype=np.float64),
            channel_radii=np.array([a]),
            l=np.array([0]),
        )
        channels = sys.build_channels(E)
        solver = LagrangeRMatrixSolver(60, 1, sys, E, channels)
        R, S, _ = solver.solve(ints, channels)
        deltaa, attena = delta(S)
        delta_grid[i] = deltaa + 1.0j * attena

    plt.plot(a_grid, np.real(delta_grid), label=r"$\mathfrak{Re}\,\delta_l$")
    plt.plot(a_grid, np.imag(delta_grid), label=r"$\mathfrak{Im}\,\delta_l$")
    plt.legend()
    plt.xlabel("channel radius [fm]")
    plt.ylabel(r"$\delta_l$ [degrees]")
    plt.show()


def local_interaction_example():
    E = 14.1
    nodes_within_radius = 5

    sys = ProjectileTargetSystem(
        reduced_mass=np.array([939.0]),
        channel_radii=np.array([nodes_within_radius * (2 * np.pi)]),
        l=np.array([1]),
        Ztarget=40,
        Zproj=1,
    )

    ch = sys.build_channels(E)

    # Lagrange-Mesh
    solver_lm = LagrangeRMatrixSolver(100, 1, sys, ecom=E)

    # Woods-Saxon potential parameters
    V0 = 60  # real potential strength
    W0 = 20  # imag potential strength
    R0 = 4  # Woods-Saxon potential radius
    a0 = 0.5  # Woods-Saxon potential diffuseness
    params = (V0, W0, R0, a0, sys.Zproj * sys.Ztarget, R0)

    ints = InteractionMatrix(1)
    ints.set_local_interaction(interaction, args=params)

    s_values = np.linspace(0.01, sys.channel_radii[0], 200)

    # Runge-Kutta
    sol_rk = solve_ivp(
        lambda s, y,: schrodinger_eqn_ivp_order1(
            s, y, ch[0], ints.local_matrix[0, 0], params
        ),
        ch[0].domain,
        ch[0].initial_conditions(),
        dense_output=True,
        atol=1.0e-12,
        rtol=1.0e-12,
    ).sol
    a = ch[0].domain[1]
    u_rk = sol_rk(s_values)[0]
    R_rk = sol_rk(a)[0] / (a * sol_rk(a)[1])
    S_rk = smatrix(R_rk, a, ch[0].l, ch[0].eta)

    R_lm, S_lm, x, uext_prime_boundary = solver_lm.solve(ints, ch, wavefunction=True)
    # R_lmp = u_lm(se.a) / (se.a * derivative(u_lm, se.a, dx=1.0e-6))
    u_lm = Wavefunctions(
        solver_lm, x, S_lm, uext_prime_boundary, sys.incoming_weights, ch
    ).uint()[0]
    u_lm = u_lm(s_values)

    R_lm = R_lm[0, 0]
    S_lm = S_lm[0, 0]

    delta_lm, atten_lm = delta(S_lm)
    delta_rk, atten_rk = delta(S_rk)

    # normalization and phase matching
    u_rk = u_rk * np.max(np.real(u_lm)) / np.max(np.real(u_rk))

    print(f"k: {ch[0].k}")
    print(f"R-Matrix RK: {R_rk:.3e}")
    print(f"R-Matrix LM: {R_lm:.3e}")
    # print(f"R-Matrix LMp: {R_lmp:.3e}")
    print(f"S-Matrix RK: {S_rk:.3e}")
    print(f"S-Matrix LM: {S_lm:.3e}")
    print(f"real phase shift RK: {delta_rk:.3e} degrees")
    print(f"real phase shift LM: {delta_lm:.3e} degrees")
    print(f"complex phase shift RK: {atten_rk:.3e} degrees")
    print(f"complex phase shift LM: {atten_lm:.3e} degrees")

    plt.plot(s_values, np.real(u_rk), "k", label="Runge-Kutta")
    plt.plot(s_values, np.imag(u_rk), ":k")

    plt.plot(s_values, np.real(u_lm), "r", label="Lagrange-Legendre")
    plt.plot(s_values, np.imag(u_lm), ":r")

    plt.legend()
    plt.xlabel(r"$r$ [fm]")
    plt.ylabel(r"$u_{%d} (r) $ [a.u.]" % ch[0].l)
    plt.tight_layout()
    plt.show()


def rmse_RK_LM():
    r"""Test with simple Woods-Saxon plus coulomb without spin-orbit coupling"""

    lgrid = np.arange(0, 6 - 1, 1)
    egrid = np.linspace(0.01, 100, 10)
    nodes_within_radius = 5

    # channels are the same except for l and uncoupled
    # so just set up a single channel system. We will set
    # incident energy and l later
    sys = ProjectileTargetSystem(
        np.array([939.0]),
        np.array([nodes_within_radius * (2 * np.pi)]),
        l=np.array([0]),
        Ztarget=40,
        Zproj=1,
        nchannels=1,
    )

    # Lagrange-Mesh solver, don't set the energy
    solver_lm = LagrangeRMatrixSolver(100, 1, sys, ecom=None)

    # Woods-Saxon potential parameters
    V0 = 60  # real potential strength
    W0 = 20  # imag potential strength
    R0 = 4  # Woods-Saxon potential radius
    a0 = 0.5  # Woods-Saxon potential diffuseness
    RC = R0  # Coulomb cutoff

    params = (V0, W0, R0, a0, sys.Zproj * sys.Ztarget, RC)

    # use same interaction for all channels
    ints = InteractionMatrix(1)
    ints.set_local_interaction(interaction, args=params)

    error_matrix = np.zeros((len(lgrid), len(egrid)), dtype=complex)

    for i, e in enumerate(egrid):
        solver_lm.set_energy(e)
        for l in lgrid:
            sys.l = np.array([l])
            ch = sys.build_channels(e)
            a = ch[0].domain[1]

            # Runge-Kutta
            sol_rk = solve_ivp(
                lambda s, y,: schrodinger_eqn_ivp_order1(
                    s, y, ch[0], ints.local_matrix[0, 0], params
                ),
                ch[0].domain,
                ch[0].initial_conditions(),
                dense_output=True,
                atol=1.0e-12,
                rtol=1.0e-12,
            ).sol

            R_rk = sol_rk(a)[0] / (a * sol_rk(a)[1])
            S_rk = smatrix(R_rk, a, l, ch[0].eta)

            # Lagrange-Legendre R-Matrix
            R_lm, S_lm, uext_boundary = solver_lm.solve(ints, ch, ecom=e)

            # comparison between solvers
            delta_lm, atten_lm = delta(S_lm)
            delta_rk, atten_rk = delta(S_rk)

            err = 0 + 0j

            if np.fabs(delta_rk) > 1e-12:
                err += np.fabs(delta_lm - delta_rk)

            if np.fabs(atten_rk) > 1e-12:
                err += 1j * np.fabs(atten_lm - atten_rk)

            error_matrix[l, i] = err

    lines = []
    for l in lgrid:
        (p1,) = plt.plot(egrid, np.real(error_matrix[l, :]), label=r"$l = %d$" % l)
        (p2,) = plt.plot(egrid, np.imag(error_matrix[l, :]), ":", color=p1.get_color())
        lines.append([p1, p2])

    plt.ylabel(r"$\Delta \equiv | \delta^{\rm RK} - \delta^{\rm LM} |$ [degrees]")
    plt.xlabel(r"$E$ [MeV]")

    legend1 = plt.legend(
        lines[0], [r"$\mathfrak{Re}\, \Delta$", r"$\mathfrak{Im}\, \Delta$"], loc=0
    )
    plt.legend([l[0] for l in lines], [l[0].get_label() for l in lines], loc=1)
    plt.yscale("log")
    plt.gca().add_artist(legend1)
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    channel_radius_dependence_test()
    local_interaction_example()
    rmse_RK_LM()



from numba.experimental import jitclass
from numba import float64, int64
import numpy as np

from .utils import hbarc, c, alpha, null
from .channel import ChannelData


class InteractionMatrix:
    r"""Represents the interaction potentials in each channel as numpy object arrays,
    one for local interactions and one for nonlocal
    """

    def __init__(self, nchannels: np.int64 = 1):
        r"""Initialize the InteractionMatrix

        Parameters:
            - nchannels (int) : the number of channels
        """
        self.nchannels = nchannels
        self.local_args = np.empty((self.nchannels, self.nchannels), dtype=object)
        self.nonlocal_args = np.empty((self.nchannels, self.nchannels), dtype=object)
        self.local_matrix = np.empty((self.nchannels, self.nchannels), dtype=object)
        self.nonlocal_matrix = np.empty((self.nchannels, self.nchannels), dtype=object)
        self.nonlocal_symmetric = np.ones((self.nchannels, self.nchannels), dtype=bool)

        # initialize local interaction to 0's
        for i in range(self.nchannels):
            for j in range(self.nchannels):
                self.local_matrix[i, j] = null
                self.local_args[i, j] = tuple()

    def set_nonlocal_interaction(
        self,
        interaction,
        i: np.int64 = 0,
        j: np.int64 = 0,
        is_symmetric=True,
        args=None,
    ):
        self.nonlocal_matrix[i, j] = interaction
        self.nonlocal_symmetric[i, j] = is_symmetric
        self.nonlocal_args[i, j] = args

    def set_local_interaction(
        self, interaction, i: np.int64 = 0, j: np.int64 = 0, args=None
    ):
        self.local_matrix[i, j] = interaction
        self.local_args[i, j] = args


system_spec = [
    ("reduced_mass", float64[:]),
    ("channel_radii", float64[:]),
    ("l", int64[:]),
    ("Ztarget", float64),
    ("Zproj", float64),
    ("nchannels", int64),
    ("level_energies", float64[:]),
    ("incoming_weights", float64[:]),
]


@jitclass(system_spec)
class ProjectileTargetSystem:
    def __init__(
        self,
        reduced_mass: np.array,
        channel_radii: np.array,
        l: np.array = None,
        Ztarget: np.float64 = 0,
        Zproj: np.float64 = 0,
        nchannels: np.int64 = 1,
        level_energies: np.array = None,
        incoming_weights: np.array = None,
    ):
        self.reduced_mass = reduced_mass
        self.channel_radii = channel_radii
        self.l = l
        self.Ztarget = Ztarget
        self.Zproj = Zproj
        self.nchannels = nchannels

        if level_energies is None:
            level_energies = np.zeros(self.nchannels)

        self.level_energies = level_energies

        if incoming_weights is None:
            incoming_weights = np.zeros(self.nchannels)
            incoming_weights[0] = 1

        self.incoming_weights = incoming_weights

        assert channel_radii.shape == (nchannels,)
        assert l.shape == (nchannels,)
        assert level_energies.shape == (nchannels,)
        assert incoming_weights.shape == (nchannels,)

    def k(self, ecom):
        r"""
        Wavenumber in each channel
        """
        return np.sqrt(2 * (ecom - self.level_energies) * self.reduced_mass) / hbarc

    def velocity(self, ecom):
        return np.sqrt(2 * hbarc * self.k(ecom) / self.reduced_mass) * c

    def eta(self, ecom):
        r"""
        Sommerfield parameter in each channel
        """
        k = self.k(ecom)
        return (alpha * self.Zproj * self.Ztarget) * self.reduced_mass / (hbarc * k)

    def build_channels(self, ecom):
        k = self.k(ecom)
        eta = self.eta(ecom)
        channels = list()
        for i in range(self.nchannels):
            channels.append(
                ChannelData(
                    self.l[i],
                    self.reduced_mass[i],
                    self.channel_radii[i],
                    ecom - self.level_energies[i],
                    k[i],
                    eta[i],
                )
            )

        return channels



import numpy as np
from numba import njit
from .utils import alpha, hbarc


@njit
def woods_saxon_potential(r, *params):
    V, W, R, a = params
    return -(V + 1j * W) / (1 + np.exp((r - R) / a))


@njit
def woods_saxon_prime(r, *params):
    """derivative of the Woods-Saxon potential w.r.t. $r$"""
    V, W, R, a = params
    return -1 * (V + 1j * W) / a * np.exp((r - R) / a) / (1 + np.exp((r - R) / a)) ** 2


@njit
def surface_peaked_gaussian_potential(r, *params):
    V, W, R, a = params
    return -(V + 1j * W) * np.exp(-((r - R) ** 2) / (2 * np.pi * a) ** 2)


@njit
def coulomb_charged_sphere(r, zz, r_c):
    return zz * alpha * hbarc * regular_inverse_r(r, r_c)


@njit
def yamaguchi_potential(r, rp, *params):
    """
    non-local potential with analytic s-wave phase shift; Eq. 6.14 in [Baye, 2015]
    """
    W0, beta, alpha = params
    return W0 * 2 * beta * (beta + alpha) ** 2 * np.exp(-beta * (r + rp))


@njit
def regular_inverse_r(r, r_c):
    if isinstance(r, float):
        return 1 / (2 * r_c) * (3 - (r / r_c) ** 2) if r < r_c else 1 / r
    else:
        ii = np.where(r <= r_c)[0]
        jj = np.where(r > r_c)[0]
        return np.hstack([1 / (2 * r_c) * (3 - (r[ii] / r_c) ** 2), 1 / r[jj]])


@njit
def yamaguchi_swave_delta(k, *params):
    """
    analytic k * cot(phase shift) for yamaguchi potential; Eq. 6.15 in [Baye, 2015]
    """
    _, a, b = params
    d = 2 * (a + b) ** 2

    kcotdelta = (
        a * b * (a + 2 * b) / d
        + (a**2 + 2 * a * b + 3 * b**2) * k**2 / (b * d)
        + k**4 / (b * d)
    )

    delta = np.rad2deg(np.arctan(k / kcotdelta))
    return delta



import numpy as np
import jitr
from numba import njit


@njit
def interaction(r, *args):
    (V0, W0, R0, a0, zz, r_c) = args
    return jitr.woods_saxon_potential(r, V0, W0, R0, a0) + jitr.coulomb_charged_sphere(
        r, zz, r_c
    )


energy_com = 26  # MeV
nodes_within_radius = 5

# initialize system and description of the channel (elastic) under consideration
sys = jitr.ProjectileTargetSystem(
    np.array([939.0]),
    np.array([nodes_within_radius * (2 * np.pi)]),
    l=np.array([0]),
    Ztarget=40,
    Zproj=1,
    nchannels=1,
)
ch = sys.build_channels(energy_com)

# initialize solver for single channel problem with 40 basis functions
solver = jitr.LagrangeRMatrixSolver(40, 1, sys)

# use same interaction for all channels
interaction_matrix = jitr.InteractionMatrix(1)
interaction_matrix.set_local_interaction(interaction)

# Woods-Saxon and Coulomb potential parameters
V0 = 60  # real potential strength
W0 = 20  # imag potential strength
R0 = 4  # Woods-Saxon potential radius
a0 = 0.5  # Woods-Saxon potential diffuseness
RC = R0  # Coulomb cutoff
params = (V0, W0, R0, a0, sys.Zproj * sys.Ztarget, RC)

# set params
interaction_matrix.local_args[0, 0] = params

# run solver
R, S, uext_boundary = solver.solve(interaction_matrix, ch, energy_com)

# get phase shift in degrees
delta, atten = jitr.delta(S[0, 0])
print(f"phase shift: {delta:1.3f} + i {atten:1.3f} [degrees]")



import numpy as np
from numba import njit
from numba.experimental import jitclass
from mpmath import coulombf, coulombg
import scipy.special as sc
from scipy.interpolate import interp1d
from scipy.misc import derivative

alpha = 1.0 / 137.0359991  # dimensionless fine structure constant
hbarc = 197.3269804  # hbar*c in [MeV femtometers]
c = 2.99792458e23  # fm/s


@njit
def complex_det(matrix: np.array):
    d = np.linalg.det(matrix @ np.conj(matrix).T)
    return np.sqrt(d)


@njit
def block(matrix: np.array, block, block_size):
    """
    get submatrix with coordinates block from matrix, where
    each block is defined by block_size elements along each dimension
    """
    i, j = block
    n, m = block_size
    return matrix[i * n : i * n + n, j * m : j * m + m]


@njit
def Gamow_factor(l, eta):
    r"""This returns the... Gamow factor.
    See [Wikipedia](https://en.wikipedia.org/wiki/Gamow_factor).

    Parameters:
        l (int): angular momentum
        eta (float): Sommerfeld parameter (see
            [Wikipedia](https://en.wikipedia.org/wiki/Sommerfeld_parameter))

    Returns:
        C_l (float): Gamow factor

    """
    if eta == 0.0:
        if l == 0:
            return 1
        else:
            return 1 / (2 * l + 1) * Gamow_factor(l - 1, 0)
    elif l == 0:
        return np.sqrt(2 * np.pi * eta / (np.exp(2 * np.pi * eta) - 1))
    else:
        return np.sqrt(l**2 + eta**2) / (l * (2 * l + 1)) * Gamow_factor(l - 1, eta)


@njit
def second_derivative_op(s, channel, interaction, args=()):
    r"""second derivative operator of reduced, scaled radial Schrodinger equation"""
    return (
        eval_scaled_interaction(s, interaction, channel, args)
        + channel.l * (channel.l + 1) / s**2
        - 1.0
    )


@njit
def schrodinger_eqn_ivp_order1(s, y, channel, interaction, args=()):
    r"""
    callable for scipy.integrate.solve_ivp; converts SE to
    2 coupled 1st order ODEs
    """
    u, uprime = y
    return [uprime, second_derivative_op(s, channel, interaction, args) * u]


class FreeAsymptotics:
    r"""For neutral particles, one may desired to explicitly pass in the type FreeAsymptotics
    into H_plus, H_minus, etc., for speed, as, while Coulomb functions reduce to the spherical
    Bessels for neutral particles, the arbitrary precision implementation in mpmath will be much
    slower than this
    """

    def __init__():
        pass

    @staticmethod
    def F(s, l, _=None):
        """
        Bessel function of the first kind.
        """
        return s * sc.spherical_jn(l, s)

    @staticmethod
    def G(s, l, _=None):
        """
        Bessel function of the second kind.
        """
        return -s * sc.spherical_yn(l, s)


class CoulombAsymptotics:
    @staticmethod
    def F(s, l, eta):
        """
        Coulomb function of the first kind.
        """
        return np.complex128(coulombf(l, eta, s))

    @staticmethod
    def G(s, l, eta):
        """
        Coulomb function of the second kind.
        """
        return np.complex128(coulombg(l, eta, s))


def H_plus(s, l, eta, asym=CoulombAsymptotics):
    """
    Hankel/Coulomb-Hankel function of the first kind (outgoing).
    """
    return asym.G(s, l, eta) + 1j * asym.F(s, l, eta)


def H_minus(s, l, eta, asym=CoulombAsymptotics):
    """
    Hankel/Coulomb-Hankel function of the second kind (incoming).
    """
    return asym.G(s, l, eta) - 1j * asym.F(s, l, eta)


def H_plus_prime(s, l, eta, dx=1e-6, asym=CoulombAsymptotics):
    """
    Derivative of the Hankel function (first kind) with respect to s.
    """
    return derivative(lambda z: H_plus(z, l, eta, asym), s, dx=dx)


def H_minus_prime(s, l, eta, dx=1e-6, asym=CoulombAsymptotics):
    """
    Derivative of the Hankel function (second kind) with respect to s.
    """
    return derivative(lambda z: H_minus(z, l, eta, asym), s, dx=dx)


def smatrix(Rl, a, l, eta, asym=CoulombAsymptotics):
    """
    Calculates channel S-Matrix from channel R-matrix (logarithmic
    derivative of channel wavefunction at channel radius)
    """
    return (
        H_minus(a, l, eta, asym=asym) - a * Rl * H_minus_prime(a, l, eta, asym=asym)
    ) / (H_plus(a, l, eta, asym=asym) - a * Rl * H_plus_prime(a, l, eta, asym=asym))


@njit
def delta(Sl):
    """
    returns the phase shift and attentuation factor in degrees
    """
    delta = np.log(Sl) / 2.0j  # complex phase shift in radians
    return np.rad2deg(np.real(delta)), np.rad2deg(np.imag(delta))


@njit
def null(*args):
    return 0.0


@njit
def eval_scaled_interaction(s, interaction, ch, args):
    return interaction(s / ch.k, *args) / ch.E


@njit
def eval_scaled_nonlocal_interaction(s, sp, interaction, ch, args):
    return interaction(s / ch.k, sp / ch.k, *args) / ch.E


@njit
def njit_eval_legendre(n, x):
    return sc.eval_legendre(n, x)



import numpy as np
from matplotlib import pyplot as plt

from jitr import (
    ProjectileTargetSystem,
    ChannelData,
    InteractionMatrix,
    LagrangeRMatrixSolver,
    Wavefunctions,
    woods_saxon_potential,
    coulomb_charged_sphere,
    surface_peaked_gaussian_potential,
    complex_det,
    FreeAsymptotics,
)


def coupled_channels_example(visualize=False):
    """
    3 level system example with local diagonal and transition potentials and neutral
    particles. Potentials are real, so S-matrix is unitary and symmetric
    """

    # Potential parameters
    V = 60  # real potential strength
    W = 0  # imag potential strength
    R = 4  # Woods-Saxon potential radius
    a = 0.5  # Woods-Saxon potential diffuseness
    params = (V, W, R, a)

    nodes_within_radius = 5

    system = ProjectileTargetSystem(
        reduced_mass=np.array([939.0, 939.0, 939]),
        channel_radii=np.ones(3) * nodes_within_radius * (2 * np.pi),
        nchannels=3,
        level_energies=np.array([0.0, 12.0, 20.0]),
        l=np.array(
            [0, 0, 0]
        ),  # for simplicity, consider the S-wave coupling to 3 different 0+ levels
    )

    interaction_matrix = InteractionMatrix(3)

    # diagonal potentials are just Woods-Saxons
    for i in range(system.nchannels):
        interaction_matrix.set_local_interaction(
            woods_saxon_potential, i, i, args=params
        )

    # transition potentials have depths damped by a factor compared to diagonal terms
    # and use surface peaked Gaussian form factors rather than Woods-Saxons
    transition_dampening_factor = 4
    Vt = V / transition_dampening_factor
    Wt = W / transition_dampening_factor
    params_off_diag = (Vt, Wt, R, a)

    # off diagonal potential terms
    for i in range(system.nchannels):
        for j in range(system.nchannels):
            if i != j:
                interaction_matrix.set_local_interaction(
                    surface_peaked_gaussian_potential,
                    i,
                    j,
                    args=params_off_diag,
                )

    ecom = 35
    channels = system.build_channels(ecom)
    solver = LagrangeRMatrixSolver(200, 3, system, ecom=ecom, channel_matrix=channels)

    H = solver.bloch_se_matrix(interaction_matrix, channels)

    if visualize:
        for i in range(3):
            for j in range(3):
                plt.imshow(
                    np.real(
                        solver.kernel.single_channel_bloch_se_matrix(
                            i,
                            j,
                            interaction_matrix.local_matrix[i, j],
                            None,
                            True,
                            channels[i],
                            None,
                            interaction_matrix.local_args[i, j],
                            None,
                        )
                    )
                )
                plt.xlabel("n")
                plt.ylabel("m")
                plt.colorbar()
                plt.title(f"({i}, {j})")
                plt.show()

    # get R and S-matrix, and both internal and external soln
    R, S, x, uext_prime_boundary = solver.solve(
        interaction_matrix, channels, wavefunction=True
    )
    u = Wavefunctions(
        solver,
        x,
        S,
        uext_prime_boundary,
        system.incoming_weights,
        channels,
        asym=FreeAsymptotics,
    ).uint()

    # S must be unitary
    assert np.isclose(complex_det(S), 1.0)

    s_values = np.linspace(0.05, system.channel_radii[0], 500)

    lines = []
    for i in range(system.nchannels):
        u_values = u[i](s_values)
        (p1,) = plt.plot(s_values, np.real(u_values), label=r"$n=%d$" % i)
        (p2,) = plt.plot(s_values, np.imag(u_values), ":", color=p1.get_color())
        lines.append([p1, p2])

    legend1 = plt.legend(
        lines[0], [r"$\mathfrak{Re}\, u_n(s) $", r"$\mathfrak{Im}\, u_n(s)$"], loc=3
    )
    plt.legend([l[0] for l in lines], [l[0].get_label() for l in lines], loc=1)
    plt.gca().add_artist(legend1)

    plt.xlabel(r"$s_n = k_n r$")
    plt.ylabel(r"$u (s) $ [a.u.]")
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    coupled_channels_example(visualize=False)
